import logging
import os
import sys

# LOG_PATH = os.environ.get("LOG_PATH", "logs")

# create log directory if it does not exist
# if not os.path.exists(LOG_PATH):
    # os.makedirs(LOG_PATH)

# Logging

# log_format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
log_format = logging.Formatter("%(levelname)s\t[%(name)s]\t%(message)s")

logger_ai4cehelper = logging.getLogger("ai4ce_helper")
logger_ai4cehelper.setLevel(logging.DEBUG)
logger_ai4cehelper.propagate = False

stream_handler_ai4cehelper = logging.StreamHandler(sys.stdout)
stream_handler_ai4cehelper.setFormatter(log_format)
# file_handler_ai4cehelper = logging.FileHandler(f"{LOG_PATH}/ai4ce-helper.log", encoding="utf-8")
# file_handler_ai4cehelper.setFormatter(log_format)

# logger_ai4cehelper.addHandler(file_handler_ai4cehelper)
logger_ai4cehelper.addHandler(stream_handler_ai4cehelper)



import uuid

import streamlit as st
import streamlit_mermaid as stmd
from anyio import sleep

from ._backend_calls import _backend_GET
from ._backend_calls import get_running_backend_version
from .functions import cleanup_parameters
from .functions import move_parameter_placeholder
from .functions_projects import get_list_of_all_project_infos
from .functions_projects import get_recent_project
from .logging_setup import logger_ai4cehelper


def st_project_selector(title: str = None, use_columns: bool = True, use_loaded: bool = False) -> int | None:

    projects_simple_list = get_list_of_all_project_infos()

    # col_title, col_button, col_selectbox, col_metric = st.columns([4, 1, 1.3, 1])
    col_1, col_2, col_3, col_4 = st.columns([4, 1, 1.3, 1])
    if not use_columns:
        col_title = col_button = col_selectbox = col_metric = st.columns(1)[0]
    else:
        col_title, col_button, col_selectbox, col_metric = (
            col_1,
            col_2,
            col_3,
            col_4,
        )
    

    with col_title:
        if title:
            st.title(title)

    with col_button:
        if st.button("Get recent project"):
            (
                st.session_state["project_id"],
                st.session_state["project_name"],
                st.session_state["modified"],
            ) = get_recent_project()

    # Let the user select a project from the list of projects in the project DB
    default_project = 0
    if "project_id" in st.session_state and use_loaded:
        for idx, project in enumerate(projects_simple_list):
            if project[0] == st.session_state["project_id"]:
                default_project = idx
                break
    with col_selectbox:
        selected_project = st.selectbox(
            "Project",
            projects_simple_list,
            index=None if "project_id" not in st.session_state or not use_loaded else default_project,
            placeholder="Select a project",
            format_func=lambda x: f"{x[1]}",
        )

    with col_metric:
        if selected_project is not None:
            st.session_state["project_id"] = selected_project[0]
            st.session_state["project_name"] = selected_project[1]

        if "project_id" not in st.session_state or st.session_state["project_id"] is None:
            st.metric(label="   ", value="<-- Please select")
        else:
            st.metric(
                label=f"Active project ID: {st.session_state['project_id']}",
                value=f"{st.session_state['project_name']}",
            )

    return st.session_state.get("project_id", None)


def st_system_generator_selector():
    """A function which adds a system generator selector to the top of the page.
    A system generator is the stored as part of the project in the backend.

    Suitable to be used in conjunction with st_project_selector().
    """
    # check if the project is set and exit without doing anything if not
    if "project_id" not in st.session_state:
        return None

    # get the list of system generators for the current project
    status_code, system_generators = _backend_GET(f"/projects/{st.session_state['project_id']}/sysgen/")

    # if successful display a selector for the system generators based on their IDs and names
    if status_code == 200:
        selected_sysgen = st.selectbox(
            "System Generator",
            [(sysgen["id"], sysgen["description"]) for sysgen in system_generators],
            index=None,
            placeholder="Select a system generator",
            format_func=lambda x: f"{x[1]}",
        )

        if selected_sysgen is not None:
            st.session_state["sysgen_id"] = int(selected_sysgen[0])
            st.session_state["sysgen_description"] = selected_sysgen[1]
    else:
        logger_ai4cehelper.error(f"(st_sgs) Error while loading system generators: {system_generators}")
        st.error(f"Error while loading system generators: {system_generators}")

    # After selecting a system generator, extract and present the generated designs from system_generators with the ID system_generators_id
    if "sysgen_id" not in st.session_state:
        st.stop()

    status_code, system_generator = _backend_GET(
        f"/projects/{st.session_state['project_id']}/sysgen/{st.session_state['sysgen_id']}/"
    )
    if status_code == 200:
        st.session_state["system_generator"] = system_generator
        st.session_state["generated_designs"] = system_generator.get("generated_designs")
    else:
        logger_ai4cehelper.error(f"(st_sgs) Error while loading generated designs: {system_generator}")
        st.error(f"Error while loading generated designs: {system_generator}")

    return st.session_state.get("sysgen_id", None)


def st_generated_design_selector():
    """A function which adds a generated design selector to the top of the page.
    A generated design is the stored as part of the project in the backend.

    Suitable to be used in conjunction with st_project_selector().
    """
    # check if the project is set and exit without doing anything if not
    if "project_id" not in st.session_state:
        return None

    # get the list of system generators for the current project
    status_code, system_generators = _backend_GET(f"/projects/{st.session_state['project_id']}/sysgen/")

    # if successful display a selector for the system generators based on their IDs and names
    if status_code == 200:
        selected_sysgen = st.selectbox(
            "System Generator",
            [(sysgen["id"], sysgen["description"]) for sysgen in system_generators],
            index=None,
            placeholder="Select a system generator",
            format_func=lambda x: f"{x[1]}",
        )

        if selected_sysgen is not None:
            st.session_state["sysgen_id"] = int(selected_sysgen[0])
            st.session_state["sysgen_description"] = selected_sysgen[1]
    else:
        logger_ai4cehelper.error(f"(st_gds) Error while loading system generators: {system_generators}")
        st.error(f"Error while loading system generators: {system_generators}")

    # After selecting a system generator, extract and present the generated designs from system_generators with the ID system_generators_id
    if "sysgen_id" in st.session_state:
        status_code, system_generator = _backend_GET(
            f"/projects/{st.session_state['project_id']}/sysgen/{st.session_state['sysgen_id']}/"
        )
        if status_code == 200:
            st.session_state["system_generator"] = system_generator
            st.session_state["generated_designs"] = system_generator.get("generated_designs")
        else:
            logger_ai4cehelper.error(f"(st_gds) Error while loading generated designs: {system_generator}")
            st.error(f"Error while loading generated designs: {system_generator}")

    # display the generated designs (comp_list_uids) in a selectbox
    # Only show design selector and warnings if a system generator has been selected
    if "sysgen_id" in st.session_state:
        if "generated_designs" in st.session_state and st.session_state["generated_designs"]:
            selected_generated_design = st.selectbox(
                "Generated Design",
                [(design["id"], design["comp_list_uids"]) for design in st.session_state["generated_designs"]],
                index=None,
                placeholder="Select a generated design",
                format_func=lambda x: f"Design #{x[0]}, Components: {x[1]}",
            )
            if selected_generated_design is not None:
                st.session_state["generated_design_id"] = selected_generated_design[0]
                st.session_state["generated_design_comp_list_uids"] = selected_generated_design[1]
        else:
            logger_ai4cehelper.warning(f"(st_gds) No generated designs available for the selected system generator.")
            st.warning("No generated designs available for the selected system generator.")

    # display the selected system generator and generated design
    if "sysgen_description" in st.session_state and "generated_design_comp_list_uids" in st.session_state:
        st.metric(
            label=f"Selected System Generator (ID: {st.session_state['sysgen_id']})",
            value=st.session_state["sysgen_description"],
        )
        st.metric(
            label=f"Selected Generated Design (ID: {st.session_state['generated_design_id']})",
            value=str(st.session_state["generated_design_comp_list_uids"]),
        )
    elif "sysgen_description" in st.session_state:
        st.metric(
            label="Selected System Generator",
            value=st.session_state["sysgen_description"],
        )

    return st.session_state.get("generated_design_id", None)


def st_load_components(uid_list: list[str] = None) -> dict[str, dict]:
    """Loads the defined UIDs from the backend into the session state under session_state['db']['components'][{uid}]
    If no list of UIDs is given, check the current projects system generators and iterate over all generated designs with their comp_list_uids.

    Displays a cute spinner while loading.
    """

    # exit if no data is given to act upon (selection has not been made previously)
    if "project_id" not in st.session_state and uid_list is None:
        return {}

    # create sessions state db if not exists
    if "db" not in st.session_state:
        st.session_state["db"] = {}
    if "components" not in st.session_state["db"]:
        st.session_state["db"]["components"] = {}

    # show infobox with loading message
    with st.spinner("Loading components..."):
        logger_ai4cehelper.info(f"(s_lc) Loading components, uid_list: {uid_list}")
        if uid_list is None and "project_id" in st.session_state:
            status_code, system_generators = _backend_GET(f"/projects/{st.session_state['project_id']}/sysgen/")
            for sysgen in system_generators:
                status_code, system_generator = _backend_GET(
                    f"/projects/{st.session_state['project_id']}/sysgen/{sysgen['id']}/"
                )
                if status_code == 200:
                    for design in system_generator.get("generated_designs", []):
                        for component_uid in design.get("comp_list_uids", []):
                            if component_uid not in st.session_state["db"]["components"]:
                                status_code, component = _backend_GET(f"/v2/components/{component_uid}")
                                if status_code == 200 and isinstance(component, dict):
                                    st.session_state["db"]["components"][component_uid] = component
                                    if component_uid != component.get("uid"):
                                        st.session_state["db"]["components"][component["uid"]] = component
                                    if "hash" in component:
                                        hash_uid = f"{component['uid'].rsplit('_', 1)[0]}_{component['hash']}"
                                        st.session_state["db"]["components"][hash_uid] = component
                                else:
                                    logger_ai4cehelper.error(
                                        f"(s_lc) Error loading component {component_uid}: {component}"
                                    )
                                    st.error(f"Error loading component {component_uid}: {component}")
                else:
                    logger_ai4cehelper.error(
                        f"(s_lc) Error loading system generator {sysgen['id']}: {system_generator}"
                    )
                    st.error(f"Error loading system generator {sysgen['id']}: {system_generator}")

        else:
            # Load each component by UID
            for component_uid in uid_list:
                status_code, component = _backend_GET(f"/v2/components/{component_uid}")
                if status_code == 200:
                    st.session_state["db"]["components"][component_uid] = component
                else:
                    logger_ai4cehelper.error(f"(s_lc) Error loading component {component_uid}: {component}")
                    st.error(f"Error loading component {component_uid}: {component}")

        for component_uid, component in st.session_state["db"]["components"].items():
            st.session_state["db"]["components"][component_uid] = cleanup_parameters(
                move_parameter_placeholder(component)
            )

    return st.session_state["db"]["components"]


def show_mermaid_diagram(mermaid_code: str = None):
    """
    Display a Mermaid diagram in Streamlit using streamlit-mermaid.

    Parameters:
        mermaid_code (str, optional): Mermaid diagram definition as a string. If not provided, a default simple graph is shown.

    Streamlit-Mermaid Options Reference:
        https://github.com/neka-nat/streamlit-mermaid/blob/main/streamlit_mermaid/frontend/src/MermaidViewer.tsx#L12
    """

    if not mermaid_code:
        mermaid_code = """
        graph LR
            A[Ex] -->|sam| D[(ple)]
        """
    stmd.st_mermaid(
        mermaid_code,
        show_controls=False,
        pan=True,
        zoom=True,
        key=f"mermaid_{uuid.uuid4()}",
    )


def st_show_backend_version():
    """Display the running backend version in the Streamlit app sidebar."""
    version = get_running_backend_version()
    st.sidebar.markdown(f"Backend Version: {version}")
    return version



from pathlib import Path
import warnings
import datetime
import pandas as pd
import httpx
from datetime import timezone
from typing import Literal
from ._backend_calls import _backend_GET
from ._backend_calls import _backend_POST
from ._backend_calls import _backend_PUT
from .functions import load_file, get_toml_from_url
from .logging_setup import logger_ai4cehelper
from deprecated import deprecated


def get_universal_sysarch_comp_mapping() -> dict:
    """
    Retrieves a universal mapping of system architecture and component structure for CubeSat systems.

    This function downloads the CubeSat system architecture mapping from a remote TOML file hosted on GitLab.
    It extracts all subsystem entries for CubeSat that are represented as uppercase keys (by convention, subsystem keys are all uppercase),
    and returns them in a hierarchical dictionary format suitable for universal lookups or further processing.

    Returns:
        dict: A dictionary containing the CubeSat system mapping, where each subsystem (uppercase key) maps to its respective components.

    Example:
        {
            "cubesat": {
                "ADCS": {...},
                "CDHS": {...},
                ...
            }
        }
    """
    url = "https://gitlab.com/ai4ce/AI4CE/-/raw/main/ai4ce/system_tags.toml?ref_type=heads"
    system_tags: dict = get_toml_from_url(url=url)

    response: dict = {}
    response["cubesat"] = {}

    for key, value in system_tags["CubeSat"].items():
        if key.isupper():  # subsystems are all uppercase
            response["cubesat"][key] = value
    return response

def create_new_project(project_info: dict):
    """Create a new project in the backend.
    Params:
        project_info(dict):
    Returns:
        dict: Backend version of the project.
    Raises:
        Exception: If there is an error creating the new project.
    """
    status_code, project = _backend_POST(endpoint=f"/v2/projects/", data=project_info)
    if status_code != 201:
        logger_ai4cehelper.error(f"(cnp) Error creating new project ({status_code}): {project}")
        raise Exception(f"Error creating new project ({status_code}): {project}")
    return project


def get_list_of_all_project_infos() -> list[tuple[int, str, str]]:
    # check for refactoring
    """A function to connect to the backend and get a list of all projects
    in the current project data base.

    Returns:
        tuple[int, str, str]: A list of tuples containing project ID, project name, and last modified date.
    
    Raises:
        Exception: If there is an error fetching the list of projects.
    """

    status_code, projects = _backend_GET(endpoint=f"/v2/projects/")
    project_ids: list[tuple[int, str, str]] = []
    if status_code == 200:
        for project in projects:
            project_ids.append((project["id"], project["project_name"], project["modified"]))
        return project_ids
    else:
        logger_ai4cehelper.error(f"(gloapi) Error fetching projects ({status_code}): {projects}")
        raise Exception(f"Error fetching projects ({status_code}): {projects}")


def get_project_info(project_id: int) -> dict:
    status_code, project_info = _backend_GET(endpoint=f"/v2/projects/{project_id}/")
    if status_code != 200:
        logger_ai4cehelper.error(f"(gpi) Error fetching project info ({status_code}): {project_info}")
        raise Exception(f"Error fetching project info ({status_code}): {project_info}")
    return project_info


def set_project_name(project_id: int, project_name: str) -> dict:
    """A function to set the project name of a specific project, identified by its ID.

    Params:
        project_id   (int): The project ID
        project_name (str): Name of the project/mission, eg OPS-Sat
    Returns:
        dict: Backend version of the general project info.
    Raises:
        KeyError: If the project is not found.
        Exception: If there is an error setting the project name.
    """

    data = {
        "id": project_id,
        "project_name": project_name,
    }

    # status, msg = _backend_put(endpoint=f"/v2/projects/{project_id}", data=data)
    # return (status, msg)
    status_code, response = _backend_PUT(endpoint=f"/v2/projects/{project_id}/", data=data)
    if status_code == 404:
        logger_ai4cehelper.error(f"(spn) Project with ID {project_id} not found.")
        raise KeyError(f"Project with ID {project_id} not found.")
    if status_code != 200:
        logger_ai4cehelper.error(f"(spn) Error setting project name ({status_code}): {response}")
        raise Exception(f"Error setting project name ({status_code}): {response}")
    return response


def get_recent_project() -> tuple[int, str, str]:
    """A function to get the most recently modified project from the backend.
    Returns:
        tuple[int, str, str]: A tuple containing project ID, project name, and last modified date.
    Raises:
        Exception: If there is an error fetching the recent project.
    """
    status_code, response = _backend_GET(endpoint=f"/v2/projects/recent/")
    if status_code != 200:
        logger_ai4cehelper.error(f"(grp) Error fetching recent project ({status_code}): {response}")
        raise Exception(f"Error fetching recent project ({status_code}): {response}")
    return (response["id"], response["project_name"], response["modified"])


def get_mission_orbit(project_id: int) -> dict:
    """A function to get the mission orbit info for a specific project based on its ID.
    Params:
        project_id (int): The project ID

    Returns:
        dict: Keplerian elements to define the orbit

    Raises:
        KeyError: If no mission is found for the given project ID.
        KeyError: If no orbit information is found in the mission data.
        Exception: If there is another error fetching the mission orbit.
    """

    status_code, response = _backend_GET(endpoint=f"/v2/projects/{project_id}/mission/")
    if status_code == 404:
        logger_ai4cehelper.error(f"(gmo) No mission orbit found for project {project_id}.")
        raise KeyError(f"No mission orbit found for project {project_id}.")
    if status_code != 200:
        logger_ai4cehelper.error(f"(gmo) Error fetching mission orbit ({status_code}): {response}")
        raise Exception(f"Error fetching mission orbit ({status_code}): {response}")
    if "orbit" not in response:
        logger_ai4cehelper.error(f"(gmo) No orbit information found in mission data")
        raise KeyError("No orbit information found in mission data")
    return response["orbit"]


def set_mission_orbit(project_id: int, orbit_info: dict) -> dict:
    """A function to set the orbit in the project database. Will create an empty mission parent if it does not exist.
    Params:
        project_id (int): The project ID
        orbit_info(dict): Information about the satellite's orbit

    Returns:
        dict: Backend version of the mission orbit or the status_code and error message.

    Raises:
        Exception: If there is an error creating or updating the mission orbit.
    """
    # check if project exists
    status_code, project_info = _backend_GET(endpoint=f"/v2/projects/{project_id}/")
    if status_code != 200:
        logger_ai4cehelper.error(f"(smo) Error fetching project info ({status_code}): {project_info}")
        raise Exception(f"Error fetching project info ({status_code}): {project_info}")

    status_code, mission_info = _backend_GET(endpoint=f"/v2/projects/{project_id}/mission/")
    if status_code == 404:
        logger_ai4cehelper.info(f"(smo) No mission info found for project {project_id}. Creating new mission info.")
        mission_info = {
            "mission_name": f"Mission {project_info['project_name']}",
            "orbit": orbit_info,
        }
        status_code, response = _backend_POST(endpoint=f"/v2/projects/{project_id}/mission/", data=mission_info)
        if status_code != 201:
            logger_ai4cehelper.error(f"(smo) Error creating mission info ({status_code}): {response}")
            raise Exception(f"Error creating mission info ({status_code}): {response}")
        return response
    elif status_code == 200:
        # update possible mission_info["orbit"] with new orbit_info
        logger_ai4cehelper.info(f"(smo) Mission info found for project {project_id}. Updating orbit info.")
        if "orbit" not in mission_info:
            mission_info["orbit"] = {}
        mission_info["orbit"].update(orbit_info)
        status_code, response = _backend_PUT(endpoint=f"/v2/projects/{project_id}/mission/", data=mission_info)
        if status_code != 200:
            logger_ai4cehelper.error(f"(smo) Error updating mission info ({status_code}): {response}")
            raise Exception(f"Error updating mission info ({status_code}): {response}")
        return response
    else:
        logger_ai4cehelper.error(f"(smo) Error fetching mission info ({status_code}): {mission_info}")
        raise Exception(f"Error fetching mission info ({status_code}): {mission_info}")


def get_enabled_components(nested_dict, enabled_components=None) -> list:
    # check for refactoring
    """
    Recursively traverses a nested dictionary to find and return a list of components that are enabled.

    Parameters:
        nested_dict (dict): The nested dictionary to traverse.
        enabled_components (list, optional): A list to store the names of the enabled components.
                                         Defaults to None, in which case a new list is created.

    Returns:
        list: A list of the names of the enabled components.
    """

    if enabled_components is None:
        enabled_components = []

    for key, value in nested_dict.items():
        if isinstance(value, dict):
            if (value.get("enabled") == True or value.get("Enabled") == True):
                enabled_components.append(key)
            get_enabled_components(value, enabled_components)

    return enabled_components


def traverse_and_modify(d: dict, sys_config_enabled: dict):
    # check for refactoring
    for key, value in d.items():
        if isinstance(value, dict):
            component = key
            traverse_and_modify(d=value, sys_config_enabled=sys_config_enabled)
        else:
            # This block executes only if `value` is not a dict, i.e., at the deepest level
            # if d["Enabled"] is True or d["Enable"] is True or :
            try:
                if d["Enabled"] is True:
                    sys_config_enabled.update(
                        enable_component(
                            component_name=key,
                            data=load_file(Path("data/backend_sys_default.json")),
                        )
                    )
            except KeyError as e:  # what are you doing here -.-
                print("Error: ", e)


def enable_component(component_name: str, data: dict) -> dict:
    # check for refactoring
    """Recursively search for a component in the nested dictionary from the backend
    and set 'enabled' to True if the feature name matches any key or value (case-insensitive).

    Parameters:
        data (dict): The nested dictionary to search within.
        feature_name (str): The feature name to search for, case-insensitively.

    Returns:
        dict: edited dict
    """

    for key, value in data.items():
        #  print(key, value)
        if isinstance(value, dict):
            enable_component(component_name, value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    enable_component(component_name, item)
        if key.lower() == component_name.lower() or str(value).lower() == component_name.lower():
            data["enabled"] = True
    return data


def set_sys_arch(project_id: int, sys_arch: dict) -> dict:
    """A function to set the system architecture in the project database. Updates the existing architecture if it exists.
    Params:
        project_id (int): The project ID
        sys_arch (dict): Information about the satellite's modules, which form the system architecture
    Returns:
        dict: Backend version of the system architecture.
    Raises:
        Exception: If there is an error creating or updating the system architecture.
    """

    status_code, system = _backend_GET(endpoint=f"/v2/projects/{project_id}/system/")
    if status_code == 404:
        logger_ai4cehelper.info(
            f"(ssa) No system architecture found for project {project_id}. Creating new system architecture."
        )
        status_code, response = _backend_POST(endpoint=f"/v2/projects/{project_id}/system/", data=sys_arch)
        if status_code != 201:
            logger_ai4cehelper.error(f"(ssa) Error creating system architecture ({status_code}): {response}")
            raise Exception(f"Error creating system architecture ({status_code}): {response}")
        return response
    elif status_code == 200:
        logger_ai4cehelper.info(
            f"(ssa) System architecture found for project {project_id}. Updating system architecture."
        )
        system.update(sys_arch)
        status_code, response = _backend_PUT(endpoint=f"/v2/projects/{project_id}/system/", data=system)
        if status_code != 200:
            logger_ai4cehelper.error(f"(ssa) Error updating system architecture ({status_code}): {response}")
            raise Exception(f"Error updating system architecture ({status_code}): {response}")
        return response


def get_sys_arch(project_id: int) -> dict:
    """A function to get the system configuration info for a specific project based on its ID.
    Params:
        project_id (int): The project ID

    Returns:
        dict: Backend version of the system architecture.

    Raises:
        KeyError: If no system architecture is found for the given project ID.
        Exception: If there is another error fetching the system architecture.
    """

    status_code, response = _backend_GET(endpoint=f"/v2/projects/{project_id}/system/")
    if status_code == 404:
        logger_ai4cehelper.error(f"(gsa) No system architecture found for project {project_id}.")
        raise KeyError(f"No system architecture found for project {project_id}.")
    if status_code != 200:
        logger_ai4cehelper.error(f"(gsa) Error fetching system architecture ({status_code}): {response}")
        raise Exception(f"Error fetching system architecture ({status_code}): {response}")
    return response


#  set_comp_list(project: int, not_yet_defined: dict)
# For each selected system generator, we need a place to store the corresponding found comp lists
# Every generator can produce n sets of components
@deprecated(reason="update_sys_generator() is deprecated, use set_sys_generator() instead.", version="0.10.12")
def update_sys_generator(project_id: int, sys_gen_info: dict) -> dict:
    """Deprecated: update_sys_generator is deprecated, use set_sys_generator instead."""
    warnings.warn("update_sys_generator is deprecated, use set_sys_generator instead.", DeprecationWarning)
    return set_sys_generator(project_id=project_id, sys_gen_info=sys_gen_info)


def set_sys_generator(project_id: int, sys_gen_info: dict) -> dict:
    """A function to set the system generator info in the project database. Updates the existing generator if it exists.
    Params:
        project_id (int): The project ID 
        sys_gen_info (dict): Information about the system generator
    Returns:
        dict: Backend version of the system generator.
    Raises:
        Exception: If there is an error creating or updating the system generator info.
    """
    if "id" in sys_gen_info:
        status_code, existing_sys_gen = _backend_GET(endpoint=f"/v2/projects/{project_id}/sysgen/{sys_gen_info['id']}/")
        if status_code == 200:
            logger_ai4cehelper.info(
                f"(ssg) System generator info found for project {project_id}. Updating system generator info."
            )
            existing_sys_gen.update(sys_gen_info)
            status_code, response = _backend_PUT(
                endpoint=f"/v2/projects/{project_id}/sysgen/{sys_gen_info['id']}/",
                data=existing_sys_gen,
            )
            if status_code != 200:
                logger_ai4cehelper.error(f"(ssg) Error updating system generator info ({status_code}): {response}")
                raise Exception(f"Error updating system generator info ({status_code}): {response}")
            return response
    logger_ai4cehelper.info(f"(ssg) Creating new system generator info for project {project_id}.")
    status_code, response = _backend_POST(endpoint=f"/v2/projects/{project_id}/sysgen/", data=sys_gen_info)
    if status_code != 201:
        logger_ai4cehelper.error(f"(ssg) Error creating system generator info ({status_code}): {response}")
        raise Exception(f"Error creating system generator info ({status_code}): {response}")
    return response


def set_trained_model(
    project_id: int,
    sysgen_id: int,
    environment_settings: dict = {},
    model_file: bytes = None,
) -> dict:

    if not model_file and not environment_settings:
        logger_ai4cehelper.error(f"(stm) Either environment_settings or model_file must be provided.")
        raise ValueError("Either environment_settings or model_file must be provided.")

    sys_gen = {
        "environment_settings": environment_settings,
        "id": sysgen_id,
    }

    sysgen_db = update_sys_generator(project_id=project_id, sys_gen_info=sys_gen)

    if model_file:
        status_code, response = _backend_GET(endpoint=f"/v2/projects/{project_id}/sysgen/{sysgen_id}/model/")
        if status_code == 404:
            logger_ai4cehelper.info(
                f"(stm) No trained model found for sysgen {sysgen_id} in project {project_id}. Creating new trained model."
            )
            status_code, response = _backend_POST(
                endpoint=f"/v2/projects/{project_id}/sysgen/{sysgen_id}/model/",
                data={"model_file": model_file},
            )
            if status_code != 201:
                logger_ai4cehelper.error(f"(stm) Error creating trained model ({status_code}): {response}")
                raise Exception(f"Error creating trained model ({status_code}): {response}")
            return response
        elif status_code == 200:
            logger_ai4cehelper.info(
                f"(stm) Trained model found for sysgen {sysgen_id} in project {project_id}. Updating trained model."
            )
            status_code, response = _backend_PUT(
                endpoint=f"/v2/projects/{project_id}/sysgen/{sysgen_id}/model/",
                data={"model_file": model_file},
            )
            if status_code != 200:
                logger_ai4cehelper.error(f"(stm) Error updating trained model ({status_code}): {response}")
                raise Exception(f"Error updating trained model ({status_code}): {response}")
            return response
        raise Exception(f"(stm) Unexpected status code ({status_code}): {response}")
    return sysgen_db


def get_prepared_system_generator_info(project_id: int) -> list:
    """A function to get all prepared system generators.
    Params:
        project_id (int): The project ID

    Returns:
        list: list of dictionaries with infos for the prepared system generators
    """

    status_code, response = _backend_GET(endpoint=f"/v2/projects/{project_id}/sysgen/")
    if status_code != 200:
        logger_ai4cehelper.error(f"(gpsg) Error fetching prepared system generators ({status_code}): {response}")
        raise Exception(f"Error fetching prepared system generators ({status_code}): {response}")
    return response


def set_sys_arch_defaults(system: dict, project_id: int) -> dict:
    """Translate a system configuration from a TOML file dictionary previously handled by load_file() to a format that can be uploaded to the backend.

    Parameters:
    system (dict): The system configuration in TOML format.

    Returns:
    dict: The system configuration in a format that can be uploaded to the backend.
    """

    sys_arch_example = {
        "system_type": "string",
        "satellite_class": "string",
        "OAP_W": 0,
        "requirements": {},
        "aodcs": {
            "enabled": False,
            "amount_of_entities": 0,
            "description": "string",
            "eFunctions": [],
            "type": "string",
            "components": [],
            "depending_parameters": [],
            "cmg": {
                "enabled": False,
                "amount_of_entities": 0,
                "description": "string",
                "eFunctions": [],
                "requirements": {},
            },
            "magnetorquers": {
                "enabled": False,
                "amount_of_entities": 0,
                "description": "string",
                "eFunctions": [],
                "requirements": {},
            },
        },
        "tcs": {
            "enabled": False,
            "amount_of_entities": 0,
            "description": "string",
            "eFunctions": [],
            "requirements": {},
            "coatings": {
                "enabled": False,
                "amount_of_entities": 0,
                "description": "string",
                "eFunctions": [],
                "requirements": {},
            },
            "heaters": {
                "enabled": False,
                "amount_of_entities": 0,
                "description": "string",
                "eFunctions": [],
                "requirements": {},
            },
        },
    }

    sys_config_enabled = {}
    # Loading System Configuration
    for key, value in system.items():
        if isinstance(value, dict):
            traverse_and_modify(d=value, sys_config_enabled=sys_config_enabled)

    response = set_sys_arch(project_id=project_id, sys_arch=sys_config_enabled)
    pass
    return response


def set_new_project(project_info: dict) -> tuple[int, dict]:
    """Create a new project in the backend. Updates the project if it already exists.
    Params:
        project_info(dict): Information about the project to create or update.
    Returns:
        tuple[int, dict]: Status code and backend version of the project.
    Raises:
        Exception: If there is an error creating the new project.
    """
    if "id" in project_info:
        status_code, existing_project = _backend_GET(endpoint=f"/v2/projects/{project_info['id']}/")
        if status_code == 200:
            logger_ai4cehelper.info(
                f"(snp) Project with ID {project_info['id']} found. Updating existing project."
            )
            existing_project.update(project_info)
            status_code, response = _backend_PUT(
                endpoint=f"/v2/projects/{project_info['id']}/",
                data=existing_project,
            )
            if status_code != 200:
                logger_ai4cehelper.error(f"(snp) Error updating project ({status_code}): {response}")
                raise Exception(f"Error updating project ({status_code}): {response}")
            return status_code, response
    logger_ai4cehelper.info(f"(snp) Creating new project.")
    status_code, response = _backend_POST(endpoint=f"/v2/projects/", data=project_info)
    if status_code != 201:
        logger_ai4cehelper.error(f"(snp) Error creating new project ({status_code}): {response}")
        raise Exception(f"Error creating new project ({status_code}): {response}")
    return status_code, response

def set_sysgen_time(project_id: int, sysgen_id:int, mode: Literal["start", "end", "last_update"]) -> dict:
    """A function to set the start time of a system generator in the project database.
    Params:
        project_id (int): The project ID
        sysgen_id (int): The system generator ID
        mode (Literal["start", "end", "last_update"]): The time mode to set.
    Returns:
        dict: Backend version of the system generator.
    Raises:
        Exception: If there is an error updating the system generator info.
    """
    status_code, sysgen_info_db = _backend_GET(endpoint=f"/v2/projects/{project_id}/sysgen/{sysgen_id}/")
    if status_code != 200:
        logger_ai4cehelper.error(f"(sst) Error fetching system generator info ({status_code}): {sysgen_info_db}")
        raise Exception(f"Error fetching system generator info ({status_code}): {sysgen_info_db}")

    sysgen_info =  {}
    sysgen_info[f"time_{mode}"] = datetime.datetime.now(timezone.utc).isoformat()

    status_code, response = _backend_PUT(
        endpoint=f"/v2/projects/{project_id}/sysgen/{sysgen_id}/",
        data=sysgen_info,
    )
    if status_code != 200:
        logger_ai4cehelper.error(f"(sst) Error updating system generator info ({status_code}): {response}")
        raise Exception(f"Error updating system generator info ({status_code}): {response}")
    return response


# def get_trained_model(project_id: int) -> mode: ASKYOUNES
#     pass


# def set_train_log(project_id: int, logs: ASKYOUNES) -> DB_response: ASKALEX
#     pass


# def get_train_logs(project_id: int) -> logs: ASKYOUNE   pass

def generate_mermaid_code(enabled_systems: list, rewards: dict = None, system_class: str = 'cubesat') -> str:
    """Generate mermaid code for a system architecture diagram based on enabled systems and optional rewards.

    Parameters:
        enabled_systems (list): A list of enabled system component names. 
        rewards (dict, optional): A dictionary mapping component names to their reward values.
        example_enabled_systems = ["eps", "batteries"]
    Returns:
        str: The generated mermaid code as a string.
    """
    mcode = []
    if system_class == 'cubesat':
        hierarchy = get_universal_sysarch_comp_mapping()
    subsystems = list(hierarchy[system_class].keys())
    subsystems_lower = [subsystem.lower() for subsystem in subsystems]
    components = [item for sublist in hierarchy[system_class].values() for item in sublist]
    # key_map = {
    #     "batteries": "r_bat",
    #     "solar_panels": "r_sp",
    #     "frames": "r_frame",
    #     "obc": "r_OBC",
    #     "cameras": "r_cam",
    #     "transceivers": "r_TRX",
    #     "reaction_wheels": "r_rw",
    #     "AODCS": "r_AODCS",
    #     "EPS": "r_EPS",
    #     "CDH": "r_CDH",
    #     "STRUCTURE": "r_Structure",
    #     "COM": "r_COMS",
    #     "PAYLOAD": "r_PL"
    # }
    mcode = ["graph TB"]
    mcode.append(
        "    classDef pink fill:#ff00ff,stroke:#000,stroke-width:2px;"
        " classDef blue color:#FFFFFF,fill:#055471,stroke:#000,stroke-width:2px;"
        " classDef yellow fill:#FFC443,stroke:#000,stroke-width:2px,color:#000;"
    )
    mcode.append(f"   {system_class}")
    mcode.append(f"    class {system_class} yellow;")
    mcode.append("    subgraph Subsystems")
    for subsystem in subsystems_lower:
        if subsystem in enabled_systems:
            mcode.append(f"    {subsystem.upper()} ")
            mcode.append(f"    class {subsystem.upper()} pink;")
    mcode.append("    end")
    mcode.append(f"    class Subsystems blue;")
    mcode.append("    subgraph Components")
    for component in components:
        if component in enabled_systems:
            mcode.append(f"    {component} ")
            mcode.append(f"    class {component} pink;")
    mcode.append("    end")
    mcode.append("    class Components blue;")
    #reward_cubesat = rewards["score"]["r_cubesat"]
    mcode.append(f"     {system_class} --- Subsystems")
    for key, value in hierarchy[system_class].items():
        if key.lower() in enabled_systems:
            #reward_subsystem = rewards["score"][key_map[key]]
            for component in value:
                if component in enabled_systems:
                    #reward_component = rewards["score"][key_map[component]]
                    mcode.append(f"     {key} --- {component}")
                    #mcode.append(f"     {key} --{key_map[component]}={reward_component}--- {component}")
    return "\n".join(mcode)


import inspect
import os
import time
import warnings
from pathlib import Path

import httpx
import streamlit as st
import toml
from deprecated import deprecated

from .logging_setup import logger_ai4cehelper


def warn_if_external_call():
    frame = inspect.currentframe()
    caller_frame = frame.f_back.f_back  # skip: warn_if_external_call -> _backend_GET -> caller
    
    caller_module = inspect.getmodule(caller_frame)
    external_function_name = caller_frame.f_code.co_name
    internal_function_name = frame.f_back.f_code.co_name

    if caller_module and not caller_module.__name__.startswith("ai4ce_helpers"):
        logger_ai4cehelper.warning(
            f"(wiec) Function '{external_function_name}' is calling internal backend function '{internal_function_name}' which"
            " handles errors internally and always returns "
            "(status_code, data/error_message). Make sure to handle these values."
        )
        warnings.warn(
            f"Function '{external_function_name}' is calling internal backend function '{internal_function_name}' which"
            " handles errors internally and always returns "
            "(status_code, data/error_message). Make sure to handle these values.",
            UserWarning,
            stacklevel=3,
        )


# Set backend base url, depending on whether the app is running in
# a docker container (which is most likely the unified interface.)
def check_if_backend_in_docker(PORT: int = 8000) -> str:
    """Check if the backend is running in a Docker container.
    If streamlit is started as part of docker, it will most likely have an environment variable set for the backend URL.
    If not, the backend is assumed to be running on localhost.

    Args:
        PORT (int): The port on which the backend is running.
    Returns:
        str: The URL of the backend.
    """
    if os.environ.get("BACKEND_URL"):
        logger_ai4cehelper.info(f"(cibid) Using BACKEND_URL from environment: {os.environ.get('BACKEND_URL')}")
        return f"{os.environ.get('BACKEND_URL')}"
    if Path("/.dockerenv").exists():
        logger_ai4cehelper.info(f"(cibid) Detected Docker environment. Using backend URL: http://backend:{PORT}")
        return f"http://backend:{PORT}"
    logger_ai4cehelper.warning(
        f"(cibid) Neither BACKEND_URL env variable nor Docker environment detected. Using localhost backend URL: http://localhost:{PORT}"
    )
    return f"http://localhost:{PORT}"


BACKEND_URL = check_if_backend_in_docker()


def check_backend_availability():
    """Check check every 5 seconds if the backend is available.
    Wait a max of 30 seconds before giving up. Display loading message
    during check, remove loading message once backend is available. Display
    error message if backend is not available.

    Check the session state to see if the backend is already available.
    If it is, do not check again.
    """
    logger_ai4cehelper.info("(cba) Checking backend availability...")
    backend_available = False
    timeout = 30

    backend_url = check_if_backend_in_docker().removesuffix("/api")

    # check if backend is already available
    if "backend_available" in st.session_state:
        backend_available = st.session_state.backend_available

    # display loading message
    checking = st.warning("Waiting for backend...")
    while not backend_available and timeout > 0:
        # check if backend is available
        # If it is, set backend_available to True and remove the loading message
        try:
            httpx.get(backend_url, timeout=1)
            backend_available = True

        except (httpx.TimeoutException, httpx.ConnectError, httpx.RemoteProtocolError):
            logger_ai4cehelper.debug(f"(cba) Backend not available yet at {backend_url}, retrying in 5 seconds...")
            time.sleep(5)
            st.session_state.backend_available = False

        timeout -= 5
        
    response = httpx.get(f"{backend_url}/api/v2/status/healthcheck", timeout=1)
    logger_ai4cehelper.debug(f"(cba) Exited backend availability check loop with backend_available={backend_available} and timeout={timeout}s")
    if not backend_available:
        logger_ai4cehelper.error(f"(cba) Backend is not available at {backend_url}")
        checking.empty()
        st.error(f"Backend is not available at {backend_url}")
        st.session_state.backend_available = False
        return
    else:
        if response is not None:
            if response.status_code != 200:
                logger_ai4cehelper.error(f"(cba) Backend returned status code {response.status_code} at {backend_url}")
                st.error(f"Backend returned status code {response.status_code} at {backend_url}")
                st.session_state.backend_available = False
            else:
                # remove loading message
                logger_ai4cehelper.info(f"(cba) Backend is available at {backend_url}")
                checking.empty()
                logger_ai4cehelper.info(f"(cba) Removing loading message.")
                st.session_state.backend_available = True
    


@st.cache_data(ttl=300)
def get_running_backend_version() -> str:
    """Get the latest version of the backend from the openapi.json file."""
    status_code, response = _backend_GET("/openapi.json")
    if status_code == 200 and isinstance(response, dict) and "info" in response and "version" in response["info"]:
        version = response["info"]["version"]
        if "a" in version:
            version = version.split("a")[0] + " (alpha" + version.split("a")[1] + ")"
        if "b" in version:
            version = version.split("b")[0] + " (beta" + version.split("b")[1] + ")"
        if "rc" in version:
            version = version.split("rc")[0] + " (rc" + version.split("rc")[1] + ")"
        logger_ai4cehelper.info(f"(g_rbv) Retrieved backend version: {version}")
        return version
    logger_ai4cehelper.error("(g_rbv) Could not retrieve backend version.")
    return "unknown"


api_excluded_endpoints = ["/requestpid", "/openapi.json"]


@deprecated(
    reason="This function is deprecated. Please use _backend_GET() instead.",
    category=DeprecationWarning,
)
def _backend_get(endpoint: str) -> dict | str:
    """Deprecated: use _backend_GET() instead."""
    warnings.warn(
        "This function is deprecated. Please use _backend_GET() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    logger_ai4cehelper.warning("(bc) _backend_get is deprecated. Use _backend_GET instead.")
    warn_if_external_call()
    return _backend_GET(endpoint=endpoint)[1]


def _backend_GET(
    endpoint: str,
    query_params: dict | None = None,
    headers: dict = {"accept": "application/json, application/toml"},
) -> tuple[int, dict | list | str]:
    """An internal function to make the development of get functions easier.

    Params:
        endpoint(str): the URL of the backend endpoint to post to

    Returns:
        dict: json response from the backend
    """
    warn_if_external_call()
    endpoint = f"/api{endpoint}" if endpoint not in api_excluded_endpoints else endpoint

    try:
        response = httpx.get(
            url=f"{BACKEND_URL}{endpoint}",
            headers=headers,
            follow_redirects=True,
            params=query_params,
        )

        # Raises an HTTPError if the response status code is 4xx or 5xx
        response.raise_for_status()
        if response.status_code == 200:
            return (
                response.status_code,
                (
                    toml.loads(response.text)
                    if response.headers["content-type"] == "application/toml"
                    else response.json()
                ),
            )
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 500:
            logger_ai4cehelper.error(f"(bc) Server Error (500):{e.response.text}")
            return (
                e.response.status_code,
                "Server Error: 500. Please check the logs for details",
            )
        elif e.response.status_code == 422:
            logger_ai4cehelper.error(f"(bc) Unprocessable Content (422): {e.response.text}")
            return (
                e.response.status_code,
                "A problem with the payload itself has been detected. Please check the logs for details.",
            )
        else:
            logger_ai4cehelper.error(f"(bc) Error ({e.response.status_code}): {e.response.text}")
            return (e.response.status_code, e.response.json())
    except (httpx.RequestError, httpx.RemoteProtocolError, httpx.ConnectError) as e:
        logger_ai4cehelper.error(f"(bc) An error occurred while requesting {e.request.url!r}.")
        return (500, f"Request Error: {e}")
    return (500, "An unknown error occurred.")


@deprecated(
    reason="This function is deprecated. Please use _backend_POST() instead.",
    category=DeprecationWarning,
)
def _backend_post(endpoint: str, data: dict) -> tuple[int, dict | str]:
    """This function is deprecated. Please use _backend_POST() instead."""
    warnings.warn(
        "This function is deprecated. Please use _backend_POST() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    logger_ai4cehelper.warning("(bc) _backend_post is deprecated. Use _backend_POST instead.")
    warn_if_external_call()
    return _backend_POST(endpoint=endpoint, data=data)


def _backend_POST(
    endpoint: str,
    data: dict,
    query_params: dict | None = None,
    headers: dict = {"Content-Type": "application/json", "accept": "application/json"},
) -> tuple[int, dict | list | str]:
    """An internal function to make the development of posting functions easier.

    Params:
        endpoint(str): the URL of the backend endpoint to post to
        data(dict): the concent to put into the backend

    Returns:
        dict: json response from the backend

    Catches:
        httpx.HTTPStatusError: if the response status code is 4xx or 5xx
        httpx.RequestError: if there is a problem with the request
    """
    warn_if_external_call()
    endpoint = f"/api{endpoint}" if endpoint not in api_excluded_endpoints else endpoint

    # Make the POST request
    try:
        response = httpx.post(
            url=f"{BACKEND_URL}{endpoint}",
            json=data,
            headers=headers,
            follow_redirects=True,
            params=query_params,
        )
        response.raise_for_status()
        if response.status_code == 201:
            return (response.status_code, response.json())
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 500:
            logger_ai4cehelper.error(f"(bp) Server Error (500):{e.response.text}")
            return (
                e.response.status_code,
                "Server Error: 500. Please check the logs for details",
            )
        elif e.response.status_code == 422:
            logger_ai4cehelper.error(f"(bp) Unprocessable Content (422): {e.response.text}")
            return (
                e.response.status_code,
                "A problem with the payload itself has been detected. Please check the logs for details.",
            )
        else:
            logger_ai4cehelper.error(f"(bp) Error ({e.response.status_code}): {e.response.text}")
            return (e.response.status_code, e.response.json())
    except (httpx.RequestError, httpx.RemoteProtocolError, httpx.ConnectError) as e:
        logger_ai4cehelper.error(f"(bp) An error occurred while requesting {e.request.url!r}.")
        return (500, f"Request Error: {e}")
    return (500, "An unknown error occurred.")


@deprecated(
    reason="This function is deprecated. Please use _backend_PUT() instead.",
    category=DeprecationWarning,
)
def _backend_put(endpoint: str, data: dict) -> tuple[int, dict | str]:
    """This function is deprecated. Please use _backend_PUT() instead."""
    warnings.warn(
        "This function is deprecated. Please use _backend_PUT() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    logger_ai4cehelper.warning("(bc) _backend_put is deprecated. Use _backend_PUT instead.")
    warn_if_external_call()
    return _backend_PUT(endpoint=endpoint, data=data)


def _backend_PUT(
    endpoint: str,
    data: dict,
    query_params: dict | None = None,
    headers: dict = {"Content-Type": "application/json", "accept": "application/json"},
) -> tuple[int, dict | list | str]:
    """An internal function to make the development of PUT functions/update easier.

    Params:
        endpoint(str): the URL of the backend endpoint to post to
        data(dict): the concent to put into the backend
        headers(dict | None): headers to include in the request

    Returns:
        dict: response from the backend
    """
    warn_if_external_call()
    endpoint = f"/api{endpoint}" if endpoint not in api_excluded_endpoints else endpoint

    # Make the PUT request
    try:
        response = httpx.put(
            url=f"{BACKEND_URL}{endpoint}",
            json=data,
            headers=headers,
            follow_redirects=True,
            params=query_params,
        )
        response.raise_for_status()
        if response.status_code == 200:
            return (
                response.status_code,
                (
                    toml.loads(response.text)
                    if response.headers["content-type"] == "application/toml"
                    else response.json()
                ),
            )
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 500:
            logger_ai4cehelper.error(f"(bput) Server Error (500):{e.response.text}")
            return (
                e.response.status_code,
                "Server Error: 500. Please check the logs for details",
            )
        elif e.response.status_code == 422:
            logger_ai4cehelper.error(f"(bput) Unprocessable Content (422): {e.response.text}")
            return (
                e.response.status_code,
                "A problem with the payload itself has been detected. Please check the logs for details.",
            )
        else:
            logger_ai4cehelper.error(f"(bput) Error ({e.response.status_code}): {e.response.text}")
            return (e.response.status_code, e.response.json())
    except (httpx.RequestError, httpx.RemoteProtocolError, httpx.ConnectError) as e:
        logger_ai4cehelper.error(f"(bput) An error occurred while requesting {e.request.url!r}.")
        return (500, f"Request Error: {e}")
    return (500, "An unknown error occurred.")



import warnings

from deprecated import deprecated

from .functions_components import export_components
from .functions_projects import get_mission_orbit
from .functions_projects import set_sys_arch_defaults
from .streamlit_functions import st_project_selector


@deprecated(
    reason="This function is deprecated. Please use get_mission_orbit() instead.",
    category=DeprecationWarning,
)
def get_mission_info(project_id: int) -> dict:
    """Deprecation warning: This function is deprecated. Please use get_mission_orbit() instead."""
    warnings.warn(
        "This function is deprecated. Please use get_mission_orbit() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_mission_orbit(project_id)


@deprecated(
    reason="This function is deprecated. Please use export_components() instead.",
    category=DeprecationWarning,
)
def get_all_decisions():
    """Deprecation warning: This function is deprecated. Please use export_components() instead."""
    warnings.warn(
        "This function is deprecated. Please use export_components() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return export_components()


@deprecated(
    reason="This function is deprecated. Please use st_project_selector() instead.",
    category=DeprecationWarning,
)
def title_with_project_selector():
    """Deprecation warning: This function is deprecated. Please use st_project_selector() instead."""
    warnings.warn(
        "This function is deprecated. Please use st_project_selector() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return st_project_selector()


@deprecated(
    reason="This function is deprecated. Please use st_project_selector() instead.",
    category=DeprecationWarning,
)
def load_project():
    """Deprecation warning: This function is deprecated. Please use st_project_selector() instead."""
    warnings.warn(
        "This function is deprecated. Please use st_project_selector() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return st_project_selector()


@deprecated(
    reason="This function is deprecated. Please use set_sys_arch_defaults() instead.",
    category=DeprecationWarning,
)
def translate_tomlsystem_to_backend(system: dict, project_id: int) -> dict:
    """Deprecation warning: This function is deprecated. Please use set_sys_arch_defaults() instead."""
    warnings.warn(
        "This function is deprecated. Please use set_sys_arch_defaults() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return set_sys_arch_defaults(system, project_id)



# Here is the place to but general helper functions
import ast
import json
import warnings
from pathlib import Path

import httpx
import pandas as pd
import toml

from .logging_setup import logger_ai4cehelper


def load_file(filename: Path) -> dict | pd.DataFrame:
    """
    Load data from a file. The function currently supports JSON files.

    Parameters:
    filename (Path): The path to the file.

    Returns:
    dict: The data loaded from the file if it's a JSON.
    """
    if filename.suffix == ".csv":
        df = pd.read_csv(filename, sep=",", header=0, index_col=None, na_values=["NA", "?"])
        return df

    elif filename.suffix == ".json":
        with open(filename) as file:
            return json.load(file)

    elif filename.suffix == ".toml":
        with open(filename) as file:
            return toml.load(file)

    else:
        logger_ai4cehelper.error("(lf) Unsupported file format: " + filename.suffix)
        raise ValueError("You need to add the file format to the load_file function.")


def move_parameter_placeholder(component: dict) -> dict:
    """

    Because not every parameter from the parameter_unit.toml is on the main level of each component, but rather in a placeholder dictionary called "i_parameter_placeholder",
    we will move them to the main level for uniformity

    Refer to the relevant functions in the Backend:
    https://gitlab.com/ai4ce/ai4ce-03-backend/-/blob/main/backend/v2/routers/validators.py?ref_type=heads#L501-515
    https://gitlab.com/ai4ce/ai4ce-03-backend/-/blob/main/backend/v2/schemas/components/_base_component_schema.py?ref_type=heads#L18-67

    And the parameter unit toml in the AI4CE main repo
    https://gitlab.com/ai4ce/public-info/-/raw/main/docs/parameter_unit_list.toml
    """

    for key, value in component.get("i_parameter_placeholder", {}).items():
        if key not in component or component[key] is None:
            component[key] = value
        elif isinstance(value, list):
            component[key].extend(value)
        else:
            logger_ai4cehelper.warning(f"(mpp) Found conflicting value for key '{key}': {component[key]} vs {value}")
    return component


def cleanup_parameters(component: dict) -> dict:
    """
    Aims to remove the 'i_' prefix from parameters in a component dictionary.

    Checks if a key starts with 'i_' and if the corresponding key without the prefix does not exist or is empty.
    Does not delete the original 'i_' key to preserve data integrity.
    """
    for key in list(component.keys()):
        if key.startswith("i_"):
            new_key = key[2:]  # Remove the 'i_' prefix
            if new_key not in component or component[new_key] in [None, "", [], {}]:
                component[new_key] = component[key]
            # del component[key]
    return component


def get_toml_from_url(url: str) -> dict:
    """
    Downloads and parses a TOML file from the specified URL.

    This function sends a GET request to the given URL, retrieves the TOML content,
    and parses it into a Python dictionary.

    Args:
        url (str): The URL pointing to the TOML file.

    Returns:
        dict: The parsed TOML content as a dictionary.

    Raises:
        httpx.HTTPStatusError: If the request to the given URL fails.
        toml.TomlDecodeError: If the file cannot be parsed as valid TOML.
    """
    with httpx.Client() as client:
        response = client.get(url)
        response.raise_for_status()
        return toml.loads(response.text)